// Welcome to
// __________         __    __  .__                               __
// \______   \_____ _/  |__/  |_|  |   ____   ______ ____ _____  |  | __ ____
//  |    |  _/\__  \\   __\   __\  | _/ __ \ /  ___//    \\__  \ |  |/ // __ \
//  |    |   \ / __ \|  |  |  | |  |_\  ___/ \___ \|   |  \/ __ \|    <\  ___/
//  |________/(______/__|  |__| |____/\_____>______>___|__(______/__|__\\_____>
//
// This file can be a nice home for your Battlesnake logic and helper functions.
//
// To get you started we've included code to prevent your Battlesnake from moving backwards.
// For more info see docs.battlesnake.com

use log::info;
use rand::seq::SliceRandom;
use serde_json::{json, Value};

use crate::{Action, Battlesnake, Board, Direction, Game, MAX_HEALTH};

impl Battlesnake {
    fn possible_actions_without_heads<'a>(
        &'a self,
        game: &'a Game,
        board: &'a Board,
    ) -> impl Iterator<Item = Direction> + 'a {
        enum_iterator::all::<Direction>()
            .filter(|direction| {
                // filter out directions that would go outside the field
                let target = self.head.move_to(*direction);
                (0..board.width).contains(&target.x) && (0..board.height).contains(&target.y)
            })
            .filter(|direction| {
                let target = self.head.move_to(*direction);
                // don't collide with any snake
                !board
                    .snakes
                    .iter()
                    .filter(|snake| {
                        // filter out snakes that are in our squad if body collisions are allowed
                        !(game.ruleset.settings.squad.allow_body_collisions
                            && self.squad == snake.squad
                            && self.id != snake.id)
                    })
                    .flat_map(|snake| {
                        // get all coordinates of the snake body. The tail of the body can be ignored
                        // if the snake hasn't just eaten, as it will move out of the way
                        let has_eaten = snake.health == MAX_HEALTH;
                        snake.body[..snake.body.len() - usize::from(!has_eaten)].iter()
                    })
                    .any(|&coord| coord == target)
            })
    }

    #[must_use]
    pub fn possible_actions(&self, game: &Game, board: &Board) -> Vec<Direction> {
        self.possible_actions_without_heads(game, board)
            .filter(|direction| {
                // don't go into spots where a bigger snake must go
                let target = self.head.move_to(*direction);
                !board
                    .snakes
                    .iter()
                    .filter(|snake| {
                        // The other snake must be bigger than we are
                        snake.length > self.length
                    })
                    .filter_map(|snake| {
                        // get all snakes movement options
                        let actions = snake
                            .possible_actions_without_heads(game, board)
                            .collect::<Vec<_>>();
                        // only snakes that have a single option
                        match actions[..] {
                            [direction] => Some(snake.head.move_to(direction)),
                            _ => None,
                        }
                    })
                    .any(|coord| coord == target)
            })
            .collect()
    }
}

// info is called when you create your Battlesnake on play.battlesnake.com
// and controls your Battlesnake's appearance
// TIP: If you open your Battlesnake URL in a browser you should see this data
pub fn info() -> Value {
    info!("INFO");

    json!({
        "apiversion": "1",
        "author": "der-informatiker",
        "color": "#00FFEE",
        "head": "smart-caterpillar",
        "tail": "mouse",
    })
}

// start is called when your Battlesnake begins a game
pub fn start(_game: &Game, _turn: i32, _board: &Board, _you: &Battlesnake) {
    info!("GAME START");
}

// end is called when your Battlesnake finishes a game
pub fn end(_game: &Game, _turn: i32, _board: &Board, _you: &Battlesnake) {
    info!("GAME OVER");
}

// move is called on every turn and returns your next move
// Valid moves are "up", "down", "left", or "right"
// See https://docs.battlesnake.com/api/example-move for available data
pub fn get_move(game: &Game, turn: i32, board: &Board, you: &Battlesnake) -> Option<Action> {
    let actions = you.possible_actions(game, board);
    if actions.is_empty() {
        return None;
    }

    // Choose a random move from the safe ones
    let chosen = actions.choose(&mut rand::thread_rng())?;

    info!("DIRECTION {}: {:?}", turn, chosen);
    Some(Action {
        r#move: *chosen,
        shout: None,
    })
}
